/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/module.h>
#define CREATE_TRACE_POINTS
#include "debug.h"

#ifdef ERRINJ
#include <linux/random.h>
#ifndef HAVE_PRANDOM
#define prandom_u32 random32
#endif
#endif

static int log_levels[LOG_MAX];

#ifdef ERRINJ
int stcvsm_error_inject(unsigned int type, const char *file, unsigned int line)
{
	static u32 probability[ERRINJ_NUM_TYPES] = {
		0x0000ffff,
		0x00ffffff,
		0x000fffff,
		0x000fffff,
		0x000fffff,
		0x000fffff,
		0x00ffffff,
		0x00ffffff,
	};

	/* TODO:  This is a hack: error injectors are only dynamically turned on if the "DEBUG"
	 * class of logs are logged at TRACE.  It is necessary to be able to dynamically control
	 * error injectors, so that test harnesses can reliably make progress at times.  It would
	 * cleaner to create a separate ioctl to set various error injector probabilities, or
	 * disable the entire set.
	 */
	if (log_levels[LOG_DEBUG] < 8)
		return 0;
	if (type < ERRINJ_NUM_TYPES) {
		if (prandom_u32() < probability[type]) {
			stcvsm_warning(LOG_DEBUG, "Error type 0x%x injected at %s:%d\n",
					type, file, line);
			return 1;
		}
	}
	return 0;
}
#endif

void stcvsm_logger(int name, int level, const char *format, ...)
{
	int l = log_levels[name] ? log_levels[name] : LOGLEVEL_WARNING;

	if (l >= level) {
		va_list args;

		va_start(args, format);
		vprintk(format, args);
		va_end(args);
	}
}

#ifdef RUNTIME_LOG
void set_log_level(int type, int level)
{
	if (level < LOGLEVEL_ERR)
		level = LOGLEVEL_ERR;
	else if (level > LOGLEVEL_TRACE)
		level = LOGLEVEL_TRACE;
	if (type >= 0 && type < LOG_MAX) {
		stcvsm_notice(LOG_DEBUG, "Setting log type %d to level %d\n", type, level);
		log_levels[type] = level;
	}
}
#endif
