/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#ifndef STCVSM_EXPORT_H
#define STCVSM_EXPORT_H

#include "portability.h"

struct exported_filter {
	int major;
	int minor;
#ifdef HAVE_BDOPS_SUBMIT_BIO
	const struct block_device_operations *orig_fops;
#else
	submit_bio_fn *orig_reqfn;
#endif
};

struct exported_state {
	u32 stcvsm_version;  /* major minor patch build */
	u32 num_vdiffs;
	u32 reserved1;
	u32 reserved2;
	int (*stcvsm_ioctl_vdiff)(dev_t dev, unsigned int cmd, void *param);
	struct exported_filter filters[1];
};

/** Manages the sending end of the upgrade transaction, which is used to transfer incremental
 * tracking state and ownership to another driver.  While the transaction is open, the driver may
 * not create or destroy vdiff or snapshot instances.  If the transaction completes successfully,
 * the driver remains quiesced and should be unloaded.
 *
 * 0: start transaction; return is the exported state, or NULL on error
 * 1: commit transaction; driver has lost ownership of all submit_bio_fn
 * 2: cancel transaction
 */
void *stcvsm_export_vdiff(unsigned int cmd);

#endif
