/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

// Some typedefs for compatibility with sbos data structures
typedef struct _OS_FILE
{
    int fd;
    char* name;
} *OS_FILE;
typedef OS_FILE *POS_FILE;
typedef uint64_t LARGE_INTEGER;
typedef unsigned char GUID[16];

// Structure for IOCTL_VSNAP_VDIFF_RESET (reanimation mode)
typedef struct _VSNAP_VDIFF_REANIMATE_INPUT_BUFFER
{
    // New GUID
    GUID NewGuid;
    // New file GUID
    GUID NewFileGuid;
    // New reset count
    uint32_t NewResetCount;
} VSNAP_VDIFF_REANIMATE_INPUT_BUFFER, *PVSNAP_VDIFF_REANIMATE_INPUT_BUFFER;

typedef struct _MAJOR_MINOR
{
    int major;
    int minor;
} MAJOR_MINOR, *PMAJOR_MINOR;
#define VDIFF_RESET_SNAPSHOT 0
#define VDIFF_RESET_GUID     1
#define VDIFF_RESET_REAN     2
typedef struct _VSNAP_IOCTL_REANIMATE_BUFF
{
    int major;
    int minor;
    int flag;
    VSNAP_VDIFF_REANIMATE_INPUT_BUFFER buf;
} VSNAP_IOCTL_REANIMATE_BUFF, *PVSNAP_IOCTL_REANIMATE_BUFF;

typedef struct _VSNAP_IOCTL_NATIVE_SUPP_BUFF
{
    MAJOR_MINOR src;
    MAJOR_MINOR dst;
} VSNAP_IOCTL_NATIVE_SUPP_BUFF, *PVSNAP_IOCTL_NATIVE_SUPP_BUFF;

typedef struct _VSNAP_IOCTL_GUID
{
    int major;
    int minor;
    GUID Guid;
} VSNAP_IOCTL_GUID, *PVSNAP_IOCTL_GUID;
// Structure for IOCTL_VSNAP_VDIFF_GET_RANGE
typedef struct _VSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER
{
    int major;
    int minor;
    uint64_t start;
    // Number of sectors
    uint32_t NumberOfSectors;
    // Are they DIFFed or not?
    bool AreDiffed;
} VSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER, *PVSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER;
// Structure for IOCTL_SNAP_VOLUME_EX
typedef struct _VSNAP_SNAP_INPUT_BUFFER_EX
{
    int major;
    int minor;
    int cow_major;
    int cow_minor;
    // TRUE if the snap-shot must be read-only
    bool ReadOnlySnapShot;
    // TRUE if files opened for backup intent must be reparsed
    bool ReparseOnBackupOpen;
	// TRUE if store is to use redirection bitmap
	bool UseRedirectBitmap;
    // TRUE if use B-tree for a table, FALSE if compressed table
    bool UseBTree;
    // Timeout in milliseconds used to wait till all IO will pause
    // The general idea is that snapping can be done only if there was no IOs
    //  during Timeout time, the start of this time interval is the last noticed IO and
    //  can precede the snap IOCTL
    // The only exception is zero MaximumTimeout, which mean - first wait, then snap "as is"
    uint32_t Timeout;
    // Maximum timeout in milliseconds, snapping will go "as is" if waiting for this long
    // 0 if wait for Timeout once, then snap "as is"
    // Otherwise, it will wait for Timeout once more, will snap if no activity
    //  of will go "as is" if the total time spent to waiting is > MaximumTimeout
    uint32_t MaximumTimeout;
    // Must be TRUE if this snap is for hot-rollback, otherwise, hot-rollback prepare will not
    //  work
    // Volume can be unsnapped in a usual simple way (not HR prepare) even if TRUE, this is
    //  HR cancel
    // If this is TRUE, then UseBTree is irrelevant, and the driver behaves as if this flag
    //  was TRUE
    // Cannot be TRUE for MV snaps
    // Field is options, the input buffer length can possibly not include it
    bool ForHotRollback;
    // VDIFF is not touched by snap/unsnap activity at all if this is TRUE
    // Otherwise, snap/unsnap will cause VDIFF updates as described above
    bool DoNotTouchVdiff;
    // TRUE if range tables must be used instead of usual bitmaps to reduce the memory
    //  footprint
    bool UseRangeTables;
#ifdef RWE
    // TRUE if use redundant write elimination logic
    bool EliminateRedundantWrites;
#endif
} VSNAP_SNAP_INPUT_BUFFER_EX, *PVSNAP_SNAP_INPUT_BUFFER_EX;

typedef struct _SNAPSHOT_NAME
{
    int major;
    int minor;
    int path_size;
    char path[32];  /* see DISK_NAME_LEN in genhd.h */
} SNAPSHOT_NAME_BUF, *PSNAPSHOT_NAME_BUF;

typedef struct _STCVSM_LOG_LEVEL_DATA
{
    int type;
    int level;
} STCVSM_LOG_LEVEL_DATA, *PSTCVSM_LOG_LEVEL_DATA;
// Structure for IOCTL_VSNAP_SET_ADD_DEV
#ifdef __cplusplus
typedef struct _VSNAP_IOCTL_SET_DEV : public VSNAP_SNAP_INPUT_BUFFER_EX
{
    GUID SetGuid;
    uint32_t SetTimeout;
    void *hEvent;
} VSNAP_IOCTL_SET_DEV, *PVSNAP_IOCTL_SET_DEV;
#else
typedef struct _VSNAP_IOCTL_SET_DEV
{
    VSNAP_SNAP_INPUT_BUFFER_EX; // requires -fms-extensions
    GUID SetGuid;
    uint32_t SetTimeout;
    void *hEvent;
} VSNAP_IOCTL_SET_DEV, *PVSNAP_IOCTL_SET_DEV;
#endif

/* NOTE: (LDD3 Chapter 6.1)
 * _IO(type,nr) (for a command that has no argument), 
 * _IOR(type,nr,datatype) (for reading data from the driver), 
 * _IOW(type,nr,datatype) (for writing data), and 
 * _IOWR(type,nr,datatype) (for bidirectional transfers).
 */
// Define a type of ioctl
#define IOCTL_VSNAP 0xCC
// Returns the VDIFF information on the volume
// Returns all zeroes if VDIFF is not on, with OutputBuffer->VDiffActive == FALSE
// Allowed on snapshot
// Input buffer: none
// Output buffer: VSNAP_VDIFF_INFO_2
#define IOCTL_VSNAP_VDIFF_STATS_2       _IOWR(IOCTL_VSNAP, 0xc9, VSNAP_VDIFF_INFO_2)
#define IOCTL_VSNAP_VDIFF_STATS_3       _IOWR(IOCTL_VSNAP, 0xc9, VSNAP_IOCTL_INFO)
// Resets the VDIFF bitmap for a volume to all zeroes
// Fails if VDIFF is not on
// Allowed on snapshot, in this case, resets the snapshot copy only and not the main volume's
//  copy
// If GUID is specified, then the VDiffResetCount will be set to 0, and the GUID will be set
//  as new DIFF GUID. GUID specification is not allowed on snapshot
// In reanimation mode (if VSNAP_VDIFF_REANIMATE_INPUT_BUFFER is provided as input)
//  - does the same, but sets the VDiffResetCount to specified value instead of 0
// Input buffer: none or the application-invented GUID, which will be used as the DIFF GUID
//                  in reanimation mode - VSNAP_VDIFF_REANIMATE_INPUT_BUFFER
// Output buffer: none
#define IOCTL_VSNAP_VDIFF_RESET         _IOW(IOCTL_VSNAP, 0xc2, VSNAP_IOCTL_REANIMATE_BUFF)
// Sets the new DiffFileGuid value for the volume
// Allowed on snapshot
// Input buffer: new DiffFileGuid
// Output buffer: none
#define IOCTL_VSNAP_VDIFF_SET_FILE_GUID _IOW(IOCTL_VSNAP, 0xca, VSNAP_IOCTL_GUID)
// Supplement a native snapshot with Vdiff. We need to trigger a vdiff copy
// from the source to the destination after the snapshot is created.
#define IOCTL_VSNAP_VDIFF_NATIVE_ON     _IOW(IOCTL_VSNAP, 0xcb, VSNAP_IOCTL_NATIVE_SUPP_BUFF)
// Supplement a native snapshot with Vdiff. We need to trigger a vdiff off
// on the destination device. The driver should know if the snapshot succeeded
// or not
#define IOCTL_VSNAP_VDIFF_NATIVE_OFF    _IOW(IOCTL_VSNAP, 0xcc, VSNAP_IOCTL_NATIVE_SUPP_BUFF)
// Retrieves the status of the sector range for the volume
// Same as IOCTL_VSNAP_VDIFF_GET_BITMAP, but information is returned in the runlist form
//  and not bitwise form
// Reports "all sectors changed" if VDIFF is not on
// Allowed on snapshot, in this case, returns the snapshot's information and not the main
//  volume's one
// Input buffer: ULONGLONG StartingSector
// Output buffer: VSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER
// Returns the number of sectors, starting from StartingSector inclusive, with the same
//  DIFF state, and their DIFF state. The next sector after this range is guaranteed to
//  have another DIFF state, send another IOCTL to determine the next run length
// The last call (disk end) returns the number of sectors as zero
#define IOCTL_VSNAP_VDIFF_GET_RANGE     _IOWR(IOCTL_VSNAP, 0xc4, VSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER)
// Starts VDIFF on the volume
// Fails if VDIFF already on
// Not allowed on snapshot
// Input buffer: none or the application-invented GUID, which will be used as the DIFF GUID
//  Arbitrary junk from memory will be used as DIFF GUID if there is no input buffer
// Output buffer: none
#define IOCTL_VSNAP_VDIFF_START         _IOW(IOCTL_VSNAP, 0xc0, VSNAP_IOCTL_GUID)
// Stops VDIFF on the volume
// Fails if VDIFF was off
// Not allowed on snapshot
// Input buffer: none
// Output buffer: none
#define IOCTL_VSNAP_VDIFF_STOP          _IOW(IOCTL_VSNAP, 0xc1, VSNAP_IOCTL_GUID)
// Same as IOCTL_SNAP_VOLUME but provides additional parameters
// Input buffer: VSNAP_SNAP_INPUT_BUFFER_EX or VSNAP_SNAP_INPUT_BUFFER_WITH_EVENT
//  (first case is the same as second case with hEvent == NULL)
// Output buffer: none
#define IOCTL_SNAP_VOLUME_EX		    _IOW(IOCTL_VSNAP, 0xa6, VSNAP_SNAP_INPUT_BUFFER_EX)
// Input buffer: none or BOOLEAN set to FALSE for usual unsnap, BOOLEAN set to TRUE for
//  hot-rollback prepare
// Output buffer: BOOLEAN, set to TRUE if all was OK, to FALSE if this snapshot previously
//  failed and was abnormally torn down (due to store overflow, out of memory for snapshot
//  maintenance or disk IO errors)
//  For HR prepare, this BOOLEAN is set to TRUE if the last step is started just now
//  (volume is dismounted), or to FALSE if the last step will be auto-started on next FS mount
#define IOCTL_UNSNAP_VOLUME			    _IOW(IOCTL_VSNAP, 0xa1, MAJOR_MINOR)
// Input buffer: none, the length is zero to get the device name
//  (suitable for DefineDosDeviceW) and 1 to get the symlink name (in user mode, just add
//      \\?\ prefix to it and open)
// The input buffer value is reserved in this case and must be NULL
// Output buffer length must be at least sizeof(USHORT), fatal error otherwise
// If output buffer is >= sizeof(USHORT), then, if the buffer is too short for a full name,
//  at least NameLengthInBytes is returned with ERROR_MORE_DATA, so the caller will know
//  the correct buffer size
// Output buffer: USHORT NameLengthInBytes, immediately followed by the Unicode snapshot name
//  (not zero terminated) to be passed to DefineDosDeviceW
#define IOCTL_GET_SNAPSHOT_NAME_EX      _IOWR(IOCTL_VSNAP, 0xd1, SNAPSHOT_NAME_BUF)
// Log Level verbosity
#define IOCTL_SET_LOG_LEVEL             _IOWR(IOCTL_VSNAP, 0xd2, STCVSM_LOG_LEVEL_DATA)

// Adds the volume to the snapshot set, possibly creating the new snapshot set
// Must be called on a filesystem volume to be added to the set
// Input buffer: VSNAP_IOCTL_SET_DEV (snapshot set GUID must be invented by the app)
//  the structure can be shorter and not include hEvent, in this case it will be assumed to be
//  NULL
// Output buffer: none
#define IOCTL_VSNAP_SET_ADD_DEV         _IOWR(IOCTL_VSNAP, 0xb0, VSNAP_IOCTL_SET_DEV)
// "Closes" the snapshot set building - no further volume additions possible
// Closes the IO barriers on all volumes in the set and does cache flushes
// Imposes a snapshot and opens the IO barriers back
// Must be called on the control device
// Input buffer: snapshot set GUID
// Output buffer: none
#define IOCTL_VSNAP_SET_SNAP            _IOWR(IOCTL_VSNAP, 0xb1, GUID)
// Unsnaps the snapshot set and destroys it
// Must be called on the control device
// Input buffer: snapshot set GUID
// Output buffer: none
#define IOCTL_VSNAP_SET_UNSNAP          _IOWR(IOCTL_VSNAP, 0xb5, GUID)
// Returns the product info of the driver binary
// As much data is copied to the output buffer as possible
// The output buffer must be at least sizeof(uint32_t) to return TotalLength
// Input buffer: TotalLength is set to allocated size
// Output buffer: VSNAP_PRODUCT_INFO with the trailing strings
#define IOCTL_VSNAP_GET_PRODUCT_INFO    _IOWR(IOCTL_VSNAP, 0xd5, PVSNAP_PRODUCT_INFO)
