/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/*
 * Utility package - UASSERT macros
 * 64bit ready
 */
#ifndef _UASSERT_H
#define _UASSERT_H

#ifdef DEBUG

#include "stcvsm_os.h"

#define UASSERT(Cond, Msg)	\
	do { if (!(Cond)) {				\
		printk(STCVSM_EMERG "ASSERTION FAILED %s %s()-%d: '%s'!\n",	\
		__FILE__, __func__, __LINE__, (Msg));		\
		BUG();					\
		} } while (0)

#define UASSERTD(Cond, Msg, Val) \
	do { if (!(Cond)) {				\
		printk(STCVSM_EMERG "ASSERTION FAILED %s %s()-%d: '%s' [%d]!\n",	\
		__FILE__, __func__, __LINE__, (Msg), (Val));	\
		BUG();					\
		} } while (0)

#define UASSERTP(Cond, Msg, Ptr) \
	do { if (!(Cond)) {				\
		printk(STCVSM_EMERG "ASSERTION FAILED %s %s()-%d: '%s' [%p]!\n",	\
		__FILE__, __func__, __LINE__, (Msg), (Ptr));	\
		BUG();					\
		} } while (0)
#else

#define UASSERT(Cond, Msg)            do {} while (0)
#define UASSERTD(Cond, Msg, Val)      do {} while (0)
#define UASSERTP(Cond, Msg, Ptr)      do {} while (0)

#endif	/* DEBUG */

#endif	/* _UASSERT_H */
