/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/init.h>
#include <linux/module.h>

#include "debug.h"
#include "export.h"
#include "filter.h"
#include "import.h"
#include "stcvsm_version.h"

MODULE_LICENSE("GPL");
MODULE_VERSION(STCVSM_VERSION);

static unsigned long export_addr;
module_param(export_addr, ulong, 0);
MODULE_PARM_DESC(export_addr, "Address of stcvsm_export_vdiff, for dynamic driver update");

int orig_write_filter(void) { return 0; }
void cow_bdev_write_endio(void) {}
void *get_snapshot_bio(void) { return NULL; }
void put_snapshot(void) {}
void stcvsm_init_proc_filter(void) {}
void stcvsm_remove_proc_filter(void) {}

static int __init stcvsm_upgrade_init(void)
{
	int err = 0;
	void *(*stcvsm_export_vdiff)(unsigned int cmd);
	if (export_addr == 0)
		return -EINVAL;
	stcvsm_export_vdiff = (void *(*)(unsigned int cmd))export_addr;

	init_filters();

	err = stcvsm_import_vdiff(stcvsm_export_vdiff);

	return err;
}

static void __exit stcvsm_upgrade_exit(void)
{
	/* VDiff is normally torn down when the upgrade driver's state is exported.  Clean up in
	 * case the export was canceled.
	 *
	 * Do not attempt to un-quiesce the old driver if the upgrade driver's state has not been
	 * exported; the old driver may have been unloaded.
	 */
	shutdown_filters();
}

module_init(stcvsm_upgrade_init);
module_exit(stcvsm_upgrade_exit);
