/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#ifndef STCVSM_WORKQUEUE_H
#define STCVSM_WORKQUEUE_H

#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/wait.h>

/** Simple non-GPL workqueue.
 *
 * Linux's workqueue implementation is only available to GPL code (unless you only use the system's
 * generic workqueues).  This is a simpler-but-sufficient implementation (not per-CPU, no
 * CPU-binding, no delayed work, etc).
 */
struct wq {
	spinlock_t lock;
	struct list_head work;
	wait_queue_head_t more_work;
	struct task_struct *thread;
	const char *name;
	void (*cleanup_fn)(void *data);
	void *cleanup_data;
};

struct wq_work;
typedef void (*wq_work_fn)(struct wq_work *);
struct wq_work {
	struct list_head list;
	wq_work_fn fn;
	long data;
};

struct wq *global_wq(void);

/** Initializes the workqueue.
 */
void wq_init(struct wq *wq, const char *name, ...);

/** Starts the workqueue; must balance with wq_stop.
 */
int wq_start(struct wq *wq);

/** Runs the work.
 */
void wq_insert_work(struct wq *wq, struct wq_work *work);

/** Finishes queued work and then stops the workqueue.  Safe to call from within the workqueue.
 */
int wq_stop(struct wq *wq, void (*cb)(void *data), void *data);

void wq_deinit(struct wq *wq);

#endif
